# Every SM has to provide the following methods:  
#   step: input ->  output
#   currentOutput: () -> output
#   done:  () -> boolean
#   start: input or None -> None
#   legalInputs: state -> list(input)
# Error to call step if done() is true

# Define a primitive state machine by specifying transition, output,
# initial state, and termination functions
# Arguments to the PrimitiveSM initializer:
#    transitionfn: (state, input) -> state
#    outputfn: state -> output
#    initStatefn: () or input -> state 
#    terminationfn: state -> boolean  (defaults to always false)
#    legalInputs: state -> list(input)
# Provides the methods listed for every SM, as well as a variable
#  currentState: state

def falseFun(x): return False

class PrimitiveSM():
    def __init__(self, transitionfn, outputfn, initStatefn,
                 terminationfn = falseFun, legalInputs = lambda : []):
        self.transitionFunction = transitionfn
        self.outputFunction = outputfn
        self.terminationFunction = terminationfn
        self.initialStateFunction = initStatefn
        self.legalInputs = legalInputs

    def step(self, input=None):
        self.currentState = self.transitionFunction(self.currentState, input)
        return self.currentOutput()
        
    def currentOutput(self):
        return self.outputFunction(self.currentState)

    def done(self):
        return self.terminationFunction(self.currentState)

    def start(self, input = None):
        self.currentState = self.initialStateFunction(input)

class SerialSM():
    def __init__(self, sm1, sm2):
        self.m1 = sm1
        self.m2 = sm2

    def step(self, input=None):
        result = self.m2.step(self.m1.currentOutput())
        self.m1.step(input)
        if not self.m2.done():     
            return result

    def currentOutput(self):
        return self.m2.currentOutput()

    def done(self):
        return self.m1.done() or self.m2.done()

    def start(self, input = None):
        self.m1.start(input)
        self.m2.start(input)

class ParallelSM ():
    def __init__(self, sm1, sm2):
        self.m1 = sm1
        self.m2 = sm2
        self.legalInputs = \
          lambda (s1, s2): [(a1, a2) for a1 in sm1.legalInputs(s1) \
                                     for a2 in sm2.legalInputs(s2)]
        
    def step(self, (i1, i2)):
        return [self.m1.step(i1), self.m2.step(i2)]

    def done(self):
        return self.m1.done() or self.m2.done()

    def start(self, input = None):
        self.m1.start(input)
        self.m2.start(input)

class FeedbackSM ():
    def __init__(self, sm):
        self.m = sm
    
    def step(self, input=None):
        if input != None:
            print "Ignoring input to feedback machine: ", input
        return self.m.step(self.m.currentOutput())

    def currentOutput(self):
        return self.m.currentOutput()
        
    def done(self):
        return self.m.done()

    def start(self, input = None):
        self.m.start(input)

class SequentialTSM ():
    # Given a list of state machines, execute the first until it is
    # done, then execute the second, etc.
    # Assume they all have the same input space
    
    def __init__(self, smList):
        self.smList = smList
        self.n = len(smList)

    def start(self, input = None):
        self.counter = 0
        self.smList[self.counter].start(input)

    def currentOutput(self):
        return self.smList[self.counter].currentOutput()

    def step(self, input=None):
        if self.smList[self.counter].done():
            if self.counter < self.n-1:
                # Start the next guy
                self.counter = self.counter + 1
                self.smList[self.counter].start(input)
                # Step ourself
                self.step(input)
        else:
            self.smList[self.counter].step(input)
        return self.currentOutput()

    def done(self):
        return self.counter == self.n-1 and self.smList[self.counter].done() 

class StreamTSM ():
    # Given a function to call that will generate a state machines,
    # generate one, execute it until done, generate another,
    # etc.  Assume they all have the same input space
    
    def __init__(self, machineGenerator):
        self.machineGenerator = machineGenerator

    def start(self, input = None):
        self.currentMachine = self.machineGenerator()
        self.currentMachine.start(input)

    def currentOutput(self):
        return self.currentMachine.currentOutput()

    def step(self, input=None):
        self.currentMachine.step(input)
        if self.currentMachine.done():
            self.start(input)
        return self.currentOutput()

    def done(self):
        return False

class TerminatingStreamTSM (StreamTSM):
    # Just like a Stream TSM, but finishes if the termination
    # condition is true.  For now, the condition can only be applied
    # to the most recent input.
    
    def __init__(self, machineGenerator, terminationCondition):
        self.terminationCondition = terminationCondition
        StreamTSM.__init__(self, machineGenerator)

    def start(self, input = None):
        self.lastInput = input
        StreamTSM.start(self, input)

    def step(self, input = None):
        self.lastInput = input
        return StreamTSM.step(self, input)

    def done(self):
        return self.terminationCondition(self.lastInput)

class RepeatTSM ():
    # Given a terminating state machine, execute it n times
    
    def __init__(self, sm, n):
        self.sm = sm
        self.n = n

    def start(self, input = None):
        self.counter = 0
        self.sm.start(input)

    def step(self, input=None):
        result = self.sm.step(input)
        if self.sm.done():
            self.counter = self.counter + 1
            if not self.done():
                self.sm.start(input)
                result = self.sm.step(input)
        return result

    def currentOutput(self):
        return self.sm.currentOutput()

    def done(self):
        return self.counter == self.n

######################################################################
#    Stochastic outputs and state transitions
######################################################################

class PrimitiveStochasticSM (PrimitiveSM):
    # Same as a PrimitiveSM, except the output of the transition
    # function and the output function should be of type Dist, and we
    # draw from them to get the actual next state and output
    def step(self, input=None):
        self.currentState = \
                       self.transitionFunction(self.currentState,input).draw()
        return self.outputFunction(self.currentState).draw()

        
######################################################################
#    Compose a plant and a controller 
######################################################################

def simulatorSM(m1, m2):
    return FeedbackSM(SerialSM(m1, m2))

######################################################################
#   Function as a degenerate state machine
######################################################################

class FunctionSM():
    def __init__(self, f):
        self.f = f

    def step(self, input):
        return self.f(input)

    def done(self):
        return False

    def start(self, input = None):
        pass

######################################################################
#    Creating new primitive state machines from compositions of
#      other machines, by explicitly construction functions
######################################################################

def makeSerialSM(sm1, sm2):
    def transitionFunction((s1, s2), a):
        newS1 = sm1.transitionFunction(s1, a)
        if not sm1.terminationFunction(s1):
            o1 = sm1.outputFunction(newS1)
            newS2 = sm2.transitionFunction(s2, o1)
            return (newS1, newS2)
    def outputFunction((s1, s2)):
        return sm2.outputFunction(s2)
    def terminationFunction((s1, s2)):
        return sm1.terminationFunction(s1) or sm2.terminationFunction(s2)
    def initialStateFunction(input = None):
        (sm1.initialStateFunction(input), sm2.initialStateFunction(input))
    return PrimitiveSM(transitionFunction, outputFunction,
                       initialStateFunction,
                       terminationFunction)

def makeParallelSM(sm1, sm2):
    def transitionFunction((s1,s2), (a1, a2)):
        return (sm1.transitionFunction(s1, a1),
                sm2.transitionFunction(s2, a2))
    def outputFunction((s1, s2)):
        return (sm1.outputFunction(s1), sm2.outputFunction(s2))
    def terminationFunction((s1, s2)):
        return sm1.terminationFunction(s1) or sm2.terminationFunction(s2)
    def initialStateFunction(input = (None, None)):
        (i1, i2) = input
        return (sm1.initialStateFunction(i1), sm2.initialStateFunction(i2))
    def legalInputs((s1, s2)):
        return [(a1, a2) for a1 in sm1.legalInputs(s1) \
                         for a2 in sm2.legalInputs(s2)]
    return PrimitiveSM(transitionFunction, outputFunction, 
                       initialStateFunction,
                       terminationFunction,
                       legalInputs)

# State of a feedback sm is both the value on the wire and the state
# of the machine
def makeFeedbackSM(sm):
    def transitionFunction((feedback, smState), input = None):
        newSmState = sm.transitionFunction(smState, feedback)
        return (sm.currentOutput(), newSmState)
    def outputFunction(s):
        return None
    return PrimitiveSM(transitionFunction, outputFunction, 
                       lambda : (sm.currentOutput, sm.initialStateFunction()),
                       sm.terminationFunction)

# Assumes all SM's in the list have the same input set
def makeSequentialTSM(smList):
    n = len(smList)
    def transitionFunction((index, state), input):
        if smList[index].done():
            index = index + 1
            if index < n:
                state = smList[index].initialStateFunction()
        if index < n:
            return (index, smList[index].transitionFunction(state, input))
    def outputFunction((index, state)):
        return smList[index].outputFunction(state)
    def terminationFunction((index, state)):
        return index == n
    return SM(transitionFunction, outputFunction,
              lambda : (0, smList[0].initialStateFunction()),
              terminationFunction)

######################################################################
#   Run a state machine that has no input
######################################################################

def run(sm, n = 10):
    i = 0
    result = [sm.currentOutput()]
    print sm.currentOutput()
    while not sm.done() and i < n:
        i = i + 1
        out = sm.step()
        print out
        result.append(out)
    return result
        
######################################################################
#   Examples
######################################################################
## Helpful simple functions

# def identity(x): return x

# ## Counter

# def makeCounter1(initialValue):
#     return FeedbackSM(PrimitiveSM(lambda s, i: i, lambda x: x+1,
#                                   lambda : initialValue))
# def makeCounter2(initialValue):
#     return makeFeedbackSM(PrimitiveSM(lambda s, i: i, lambda x: x+1,
#                                       lambda : initialValue))

# ## Running average.  State is (n, sum).
# def makeAvgSM():
#     def avgTF((n, sum), input):
#         return (n+1, sum+input)
#     def avgOF((n, sum)):
#         if n != 0:
#             return sum/float(n)
#     return PrimitiveSM(avgTF, avgOF, lambda: (0, 0))

# ## Fib.  State is (f(n-1), f(n-2))
# def makeFibSM():
#     def fibTF((fn1, fn2), i):
#         return (fn1+fn2, fn1)
#     def fibOF((fn1, fn2)):
#         return fn1
#     return PrimitiveSM(fibTF, fibOF, lambda: (1, 1))
                     

#############################################################################
##  Get user input
#############################################################################

import sys
def makeUserPromptSM(converter = lambda x: x):
    # State of this machine is (done?, output)
    def userPromptTF(state, input):
        print "Machine's output: ", input
        print "Your input (type 'Done' to terminate): " 
        result = sys.stdin.readline()[:-1]
        return (result == 'Done', result)
    def userPromptOF((isDone, output)):
        return converter(output)
    def userPromptDone((isDone, output)):
        return isDone
    return PrimitiveSM(userPromptTF, userPromptOF,
                       lambda: (False, None),
                       userPromptDone)

#############################################################################
##  Average user input
#############################################################################

# foo = simulatorSM(makeUserPromptSM(int), makeAvgSM())

#############################################################################
##  Sequencing
#############################################################################
def makeCharSM(c):
    # State is True or False (done or not)
    return PrimitiveSM(lambda s, i: True, lambda s: c,
                       lambda: False, lambda s: s)

def makeTextSequenceSM(str):
    return SequentialTSM([makeCharSM(c) for c in str])

# squareSM = RepeatTSM(makeTextSequenceSM('forward-turn-'), 4)

